#include "MeshDrawerCoreVisitor.h"
#include <GL/gl.h>
#include <iostream>

using namespace std;

namespace X3DTK {
namespace MESH {

MeshDrawerCoreVisitor::MeshDrawerCoreVisitor()
{
  // Enter and leave functions.
  defineNewEnterFunction(&MeshDrawerCoreVisitor::enterTransform);
  defineNewEnterFunction(&MeshDrawerCoreVisitor::enterMesh);
  defineNewLeaveFunction(&MeshDrawerCoreVisitor::leaveTransform);
  
  // GlobalVariables assignation.
  globalVariables = GVManager::getInstanceOf<MeshDrawerGlobalVariables>();
}

void MeshDrawerCoreVisitor::enterTransform(Transform *T) const
{
  globalVariables->pushMatrix(T->getTransform());
  glPushMatrix();
  glMultMatrixf(T->getTransform().toFloat16());
}

void MeshDrawerCoreVisitor::leaveTransform(Transform *) const
{
  globalVariables->popMatrix();
  glPopMatrix();
}

void MeshDrawerCoreVisitor::enterMesh(Mesh *M) const
{
  globalVariables->storeMesh(M);
  int meshId = globalVariables->getCurrentMeshId();

  if (globalVariables->getSelectionMode())
    glPushName(meshId);

  bool colorPerVertex = M->data().getColorPerVertex();
  bool normalPerVertex = M->data().getNormalPerVertex();
  
  switch (globalVariables->getPrimitiveType())
  {
    case 1 : // MeshViewer::VERTICES
    {
      glColor3f(1.0, 0.9, 0.8);
      glPointSize(6.0);
      const Mesh::MVertex& vertices = M->getVertices();

      if (globalVariables->getSelectionMode())
      {
        glPushName(1);
        for (Mesh::MVertex::const_iterator v = vertices.begin(), end = vertices.end() ; v != end; ++v)
        {
          glPushName((*v)->getIndex());
          SFPoint3f point = (*v)->data().getPoint();
          glRasterPos3fv(point.f_data());
          glPopName();
        }
        glPopName();
      }
      else
      {
        glBegin(GL_POINTS);
        for (Mesh::MVertex::const_iterator v = vertices.begin(), end = vertices.end() ; v != end; ++v)
        {
          SFPoint3f point = (*v)->data().getPoint();
          glVertex3fv(point.f_data());
        }
        glEnd();
      }
      break;
    }
    
    case 2 : // MeshViewer::EDGES
    {
      glLineWidth(2.0);
      const Mesh::MEdge& edges = M->getEdges();

      if (globalVariables->getSelectionMode())
        glPushName(2);

      int index = 0;
      for (Mesh::MEdge::const_iterator e = edges.begin(), end = edges.end() ; e != end; ++e)
      {
        if (globalVariables->getSelectionMode())
          glPushName(index++);

        if ((*e)->isBoundary())
          glColor3f(0.3, 0.3, 0.9);
        else
          glColor3f(0.7, 0.7, 0.7);

        // Normal
        Mesh::Face* firstFace = (*((*e)->getLeftFaces().begin()));
         
        if (!firstFace)
          firstFace = (*((*e)->getRightFaces().begin()));
        SFVec3f normal;
        if (normalPerVertex)
          normal = (*e)->getFromVertex()->data().getNormalOfFace(firstFace);
        else
          normal = firstFace->data().getNormal();
         
        glNormal3fv(normal.f_data());
	   
        glBegin(GL_LINES);
        SFPoint3f point = (*e)->getFromVertex()->data().getPoint();
        glVertex3fv(point.f_data());
        point = (*e)->getToVertex()->data().getPoint();
        glVertex3fv(point.f_data());
        glEnd();
	    
        if (globalVariables->getSelectionMode())
          glPopName();
      }
	
      if (globalVariables->getSelectionMode())
        glPopName();
	
      break;
    }

    case 3 : // MeshViewer::FACES
    {
      const Mesh::MFace& faces = M->getFaces();
	
      if (globalVariables->getSelectionMode())
        glPushName(3);

      if (globalVariables->getOneColorPerMesh())
      {
        SFColorRGBA color = SFColorRGBA(0.1*((3+meshId*3)%10), 0.1*((5+meshId*4)%10), 0.1*((7+meshId*5)%10), 1.0);
        glColor4f(color.r, color.g, color.b, color.a);
      }
	
      for (Mesh::MFace::const_iterator f = faces.begin(), end = faces.end() ; f != end; ++f)
      {
        if (!normalPerVertex)
        {
          SFVec3f normal = (*f)->data().getNormal();
          glNormal3fv(normal.f_data());
        }

        if ((!colorPerVertex) && (!globalVariables->getOneColorPerMesh()))
        {
          SFColorRGBA color = (*f)->data().getColor();
          glColor4f(color.r, color.g, color.b, color.a);
        }
	    
	if (globalVariables->getSelectionMode())
          glPushName((*f)->getIndex());

        glBegin(GL_POLYGON);
        const Mesh::MEdge& edges = (*f)->getEdges();
        for (Mesh::MEdge::const_iterator e = edges.begin(), end = edges.end() ; e != end; ++e)
        {
          if (normalPerVertex)
          {
            SFVec3f normal = (*e)->getFromVertex()->data().getNormalOfFace(*f);
            glNormal3fv(normal.f_data());
          }
          if ((colorPerVertex) && (!globalVariables->getOneColorPerMesh()))
          {
            SFColorRGBA color = (*e)->getFromVertex()->data().getColorOfFace(*f);
            glColor4f(color.r, color.g, color.b, color.a);
          }
          
          SFPoint3f point = (*e)->getFromVertex()->data().getPoint();
          glVertex3fv(point.f_data());
        }
        glEnd();

        if (globalVariables->getSelectionMode())
          glPopName();
      }

      if (globalVariables->getSelectionMode())
        glPopName();

      break;
    }
  }

  if (globalVariables->getSelectionMode())
    glPopName();
}

}
}
