#include "libXMLLoader.h"
#include "libXMLFileElement.h"
#include "Creator.h"
#include "X3DAbstractNode.h"
#include "Scene.h"

#include <libxml/parser.h>
#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

typedef struct _X3DParseState X3DParseState;
struct _X3DParseState {
  MFAbstractNode _nodeStack;
  X3D::Scene *_scene;
  Creator *_creator;
  bool _sceneDescription;
  unsigned int _depth;
};

static void
x3d_parser_start_document(X3DParseState *state)
{
  state->_scene = 0;
  state->_sceneDescription = false;
  state->_depth = 0;
}

static void
x3d_parser_end_document(X3DParseState *state)
{
  X3DAbstractNode::DEFDict.makeDEFNonAmbiguous();
  state->_nodeStack.clear();
}

static void
x3d_parser_start_element(X3DParseState *state, const xmlChar *name, const xmlChar **attrs)
{
  if (state->_depth == 0)
  {
    libXMLFileElement e(name, attrs);

    int index;  
    if (state->_sceneDescription)
    {
      SFAbstractNode Father = state->_nodeStack.back();
      SFAbstractNode N;
      index = e.getIndexAttribute("USE");
      
      if (index != -1)
      {
        N = X3DAbstractNode::DEFDict.getNodeOfName(e.getAttribute(index));
      }  
      else  
      {
        X3D::SFNode cN = state->_creator->createFromName(e.getName());
        
        if (cN == 0)
        {
          state->_depth = 1;
          return;
        }
        cN->load(&e);
        N = cN;
      }
      index = e.getIndexAttribute("DEF");
      if (index != -1)
        X3DAbstractNode::DEFDict.insert(e.getAttribute(index), N);

      Father->setChild(N);
      state->_nodeStack.push_back(N);  
    }  
  
    if (e.getName() == "Scene")
    {
      state->_sceneDescription = true;
      state->_scene = new X3D::Scene();
      state->_nodeStack.push_back(state->_scene);
    } 
  }
  else
    ++state->_depth;
}

static void
x3d_parser_end_element(X3DParseState *state, const xmlChar *name)
{
  if (state->_sceneDescription)
  {
    if ((state->_depth == 0) && (state->_nodeStack.size() > 0))
      state->_nodeStack.pop_back();
    else
      --state->_depth;  
  }
}

static void
x3d_parser_error(X3DParseState *state, const char *msg, ...)
{
}
 
static void
x3d_parser_fatal_error(X3DParseState *state, const char *msg, ...)
{
}

static xmlSAXHandler x3d_parser = 
{
  0, /* internalSubset */
  0, /* isStandalone */
  0, /* hasInternalSubset */
  0, /* hasExternalSubset */
  0, /* resolveEntity */
  0, /* getEntity */
  0, /* entityDecl */
  0, /* notationDecl */
  0, /* attributeDecl */
  0, /* elementDecl */
  0, /* unparsedEntityDecl */
  0, /* setDocumentLocator */
  (startDocumentSAXFunc)x3d_parser_start_document, /* startDocument */
  (endDocumentSAXFunc)x3d_parser_end_document, /* endDocument */
  (startElementSAXFunc)x3d_parser_start_element, /* startElement */
  (endElementSAXFunc)x3d_parser_end_element, /* endElement */
  0, /* reference */
  0, /* characters */
  0, /* ignorableWhitespace */
  0, /* processingInstruction */
  0, /* comment */
  0, /* warning */
  (errorSAXFunc)x3d_parser_error, /* error */
  (fatalErrorSAXFunc)x3d_parser_fatal_error, /* fatalError */
};


libXMLLoader::libXMLLoader(Creator *creator)
{
  this->creator = creator;
}

libXMLLoader::~libXMLLoader()
{
  delete creator;
}

X3D::Scene *libXMLLoader::loadFile(const char *file, bool fileValidation) const
{
  SFString sfile(file);
  
  // Goes to the directory of the file.
  char *cwd = (char *)malloc(65535);
  char *value = getcwd(cwd, 65535);
  if (value == 0)
  {
    cerr << "libXMLLoader::load: unable to get current working directory!" << endl;
    return 0;
  }  
  
  if (chdir(sfile.getPath()) != 0)
  {
    cerr << "libXMLLoader::load: unable to set the current working directory to " << sfile.getPath() << "!" << endl;
    return 0;
  }
      
  X3DParseState state;
  state._creator = creator;
  
  if (xmlSAXUserParseFile(&x3d_parser, &state, file) < 0) 
  {
    cerr << "libXMLLoader::load: unable to load the file" << endl;
    return 0;
  }
  // Goes back to the initial directory.
  if (chdir(cwd) != 0)
  {
    cerr << "libXMLLoader::load: unable to set the current working directory to !" << cwd << endl;
    return 0;
  }
  
  free(cwd);
  
  X3D::Scene *scene = state._scene;
  if (scene != 0)
    scene->setFileName(file);
      
  return scene;
}

}
}
