#include "NurbsSurface.h"
#include "X3DTextureCoordinateNode.h"

#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

NurbsSurface::NurbsSurface()
  : X3DParametricGeometryNode(),
    _controlPoint(MFVec3f()),
    _texCoord    (0),
    _uTesselation(0),
    _vTesselation(0),
    _weight      (MFDouble()),
    _ccw         (true),
    _solid       (true),
    _uDimension  (0),
    _vDimension  (0),
    _uKnot       (MFDouble()),
    _vKnot       (MFDouble()),
    _uOrder      (3),
    _vOrder      (3)
{
  defineTypeName("NurbsSurface", "NURBS");
}

NurbsSurface::NurbsSurface(const MFVec3f                  &controlPoint,
                           const X3DTextureCoordinateNode *texCoord,
                           SFInt32                        uTesselation,
                           SFInt32                        vTesselation,
                           const MFDouble                 &weight,
                           SFBool                         ccw,
                           SFBool                         solid,
                           SFInt32                        uDimension,
                           SFInt32                        vDimension,
                           const MFDouble                 &uKnot,
                           const MFDouble                 &vKnot,
                           SFInt32                        uOrder,
                           SFInt32                        vOrder)
  : X3DParametricGeometryNode(),
    _controlPoint(controlPoint),
    _texCoord    ((SFNode)texCoord),
    _uTesselation(uTesselation),
    _vTesselation(vTesselation),
    _weight      (weight),
    _ccw         (ccw),
    _solid       (solid),
    _uDimension  (uDimension),
    _vDimension  (vDimension),
    _uKnot       (uKnot),
    _vKnot       (vKnot),
    _uOrder      (uOrder),
    _vOrder      (vOrder)
{
  defineTypeName("NurbsSurface", "NURBS");
}

NurbsSurface::NurbsSurface(const NurbsSurface &N)
  : X3DParametricGeometryNode(N),
    _controlPoint(N._controlPoint),
    _texCoord    (N._texCoord),
    _uTesselation(N._uTesselation),
    _vTesselation(N._vTesselation),
    _weight      (N._weight),
    _ccw         (N._ccw),
    _solid       (N._solid),
    _uDimension  (N._uDimension),
    _vDimension  (N._vDimension),
    _uKnot       (N._uKnot),
    _vKnot       (N._vKnot),
    _uOrder      (N._uOrder),
    _vOrder      (N._vOrder)
{
}

SFNode NurbsSurface::clone() const
{
  return new NurbsSurface(*this);
}

NurbsSurface::~NurbsSurface()
{
  removeParentFromChild(this, _texCoord);
}

void NurbsSurface::setControlPoint(const MFVec3f &controlPoint)
{
  _controlPoint = controlPoint;
}

void NurbsSurface::setTexCoord(const X3DTextureCoordinateNode *texCoord)
{
  removeParentFromChild(this, _texCoord);
  _texCoord = (SFNode)texCoord;
  addParentToChild(this, _texCoord);
}

void NurbsSurface::setUTesselation(SFInt32 uTesselation)
{
  _uTesselation = uTesselation;
}

void NurbsSurface::setVTesselation(SFInt32 vTesselation)
{
  _vTesselation = vTesselation;
}

void NurbsSurface::setWeight(const MFDouble &weight)
{
  _weight = weight;
}

void NurbsSurface::setCcw(SFBool ccw)
{
  _ccw = ccw;
}

void NurbsSurface::setSolid(SFBool solid)
{
  _solid = solid;
}

void NurbsSurface::setUDimension(SFInt32 uDimension)
{
  _uDimension = uDimension;
}

void NurbsSurface::setVDimension(SFInt32 vDimension)
{
  _vDimension = vDimension;
}

void NurbsSurface::setUKnot(const MFDouble &uKnot)
{
  _uKnot = uKnot;
}

void NurbsSurface::setVKnot(const MFDouble &vKnot)
{
  _vKnot = vKnot;
}

void NurbsSurface::setUOrder(SFInt32 uOrder)
{
  _uOrder = uOrder;
}

void NurbsSurface::setVOrder(SFInt32 vOrder)
{
  _vOrder = vOrder;
}

bool NurbsSurface::addChild(const SFAbstractNode &N)
{
  if (dynamic_cast<X3DTextureCoordinateNode *>(N) != 0)
  {
    if (_texCoord != 0)
      return false;
    
    setTexCoord(static_cast<X3DTextureCoordinateNode *>(N));   
    return true;    
  }
  return false;
}

bool NurbsSurface::setChild(const SFAbstractNode &N)
{
  if (dynamic_cast<X3DTextureCoordinateNode *>(N) != 0)
  {
    setTexCoord(static_cast<X3DTextureCoordinateNode *>(N));
    return true;    
  }
  
  cerr << "NurbsSurface::addChild : a node of type " << N->getTypeName() << " cannot be a child!" << endl;
  return false;
}

bool NurbsSurface::removeChild(const SFAbstractNode &N)
{
  if (_texCoord == N)
  {
    setTexCoord(0);
    return true;
  } 
  
  cerr << "NurbsSurface::removeChild : a node of type " << N->getTypeName() << " cannot be a child!" << endl;
  return false;
}

void NurbsSurface::load(const X3DFileElement *element)
{
  int index;
  index = element->getIndexAttribute("controlPoint");
  if (index != -1)
    _controlPoint = MFVec3f(element->getAttribute(index));

  index = element->getIndexAttribute("uTesselation");
  if (index != -1)
    _uTesselation = element->getAttribute(index).toInt();

  index = element->getIndexAttribute("vTesselation");
  if (index != -1)
    _vTesselation = element->getAttribute(index).toInt();

  index = element->getIndexAttribute("weight");
  if (index != -1)
    _weight = MFDouble(element->getAttribute(index));

  index = element->getIndexAttribute("ccw");
  if (index != -1)
    _ccw = (element->getAttribute(index).upper() == "TRUE");

  index = element->getIndexAttribute("solid");
  if (index != -1)
    _solid = (element->getAttribute(index).upper() == "TRUE");

  index = element->getIndexAttribute("uDimension");
  if (index != -1)
    _uDimension = element->getAttribute(index).toInt();

  index = element->getIndexAttribute("vDimension");
  if (index != -1)
    _vDimension = element->getAttribute(index).toInt();

  index = element->getIndexAttribute("uKnot");
  if (index != -1)
    _uKnot = MFDouble(element->getAttribute(index));

  index = element->getIndexAttribute("vKnot");
  if (index != -1)
    _vKnot = MFDouble(element->getAttribute(index));

  index = element->getIndexAttribute("uOrder");
  if (index != -1)
    _uOrder = element->getAttribute(index).toInt();

  index = element->getIndexAttribute("vOrder");
  if (index != -1)
    _vOrder = element->getAttribute(index).toInt();
}

SFString NurbsSurface::write() const
{
  SFString attr;
  if (!_controlPoint.empty())
    attr += " controlPoint=\"" + toSFString(_controlPoint) + "\"";
  if (_uTesselation != 0)
    attr += " uTesselation=\"" + toSFString(_uTesselation) + "\"";
  if (_vTesselation != 0)
    attr += " vTesselation=\"" + toSFString(_vTesselation) + "\"";
  if (!_weight.empty())
    attr += " weight=\"" + toSFString(_weight) + "\"";
  if (!_ccw)
    attr += " ccw=\"FALSE\"";
  if (!_solid)
    attr += " solid=\"FALSE\"";
  if (_uDimension != 0)
    attr += " uDimension=\"" + toSFString(_uDimension) + "\"";
  if (_vDimension != 0)
    attr += " vDimension=\"" + toSFString(_vDimension) + "\"";
  if (!_uKnot.empty())
    attr += " uKnot=\"" + toSFString(_uKnot) + "\"";
  if (!_vKnot.empty())
    attr += " vKnot=\"" + toSFString(_vKnot) + "\"";
  if (_uOrder != 3)
    attr += " uOrder=\"" + toSFString(_uOrder) + "\"";
  if (_vOrder != 3)
    attr += " vOrder=\"" + toSFString(_vOrder) + "\"";
      
  return attr;
}

}
}
