#include "MeshX3DMergerGeometry3DVisitor.h"
#include "IndexedFaceSet.h"
#include "Coordinate.h"
#include "Normal.h"
#include "Color.h"
#include "ColorRGBA.h"
#include "TextureCoordinate.h"

#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

MeshX3DMergerGeometry3DVisitor::MeshX3DMergerGeometry3DVisitor()
{
  globalVariables = GVManager::getInstanceOf<MeshX3DMergerGlobalVariables>();

  defineNewLeaveFunction(&MeshX3DMergerGeometry3DVisitor::leaveIndexedFaceSet);
}

MeshX3DMergerGeometry3DVisitor::~MeshX3DMergerGeometry3DVisitor()
{
}

void MeshX3DMergerGeometry3DVisitor::leaveIndexedFaceSet(IndexedFaceSet *I) const
{
  globalVariables->popParent();
  
  MESH::Mesh *MM = globalVariables->getNextMesh();
  if (MM == 0)
  {
    globalVariables->setExit(true);
    cerr << "MeshX3DMerger: cannot merge!" << endl; 
    return;
  }  
  
  // setting the parameters.
  I->setCcw(true);
  I->setNormalPerVertex(MM->data().getNormalPerVertex());
  I->setNormalPerVertex(MM->data().getColorPerVertex());
  
  // Coordinate creation.
  Coordinate *C = dynamic_cast<Coordinate *>(I->getCoord());
  if (C == 0)
  {
    C = new Coordinate();
    I->setCoord(C);
  }
  
  // filling the coordinate array.
  MFVec3f coordArray;
  const MESH::Mesh::MVertex &mv = MM->getVertices(); 
  
  // need to get the older iterator... to manage NULL references.
  // iterating the vertices.
  for (MESH::Mesh::MVertex::const_iterator itVertex = mv.begin(); itVertex != mv.end(); ++itVertex)
    coordArray.push_back(SFVec3f((*itVertex)->data().getPoint()));
  
  C->setPoint(coordArray);
  
  // Faces
  MFInt32 coordIndexArray;
  const MESH::Mesh::MFace &mf = MM->getFaces(); 
  
  for (MESH::Mesh::MFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
  {
    const MESH::Mesh::MEdge &me = (*itFace)->getEdges();
    for (MESH::Mesh::MEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
      coordIndexArray.push_back((*itEdge)->getFromVertex()->getIndex());
    coordIndexArray.push_back(-1);
  }
  
  I->setCoordIndex(coordIndexArray);

  //////////////////////////////////////////////////////////////////////////////////// 
  // normals
  
  if (MM->data().hasNormal())
  {
    Normal *N = dynamic_cast<Normal *>(I->getNormal());
    if (N == 0)
    {
      N = new Normal();
      I->setNormal(N);
    }
    
    MFVec3f normalArray;

    if (MM->data().getNormalPerVertex())
    {
      // iterating the vertices to collect the informations relative to the normals
      vector<unsigned int> normalDecal;
      for (MESH::Mesh::MVertex::const_iterator itVertex = mv.begin(); itVertex != mv.end(); ++itVertex)
      {
        normalDecal.push_back(normalArray.size());
        
        MFVec3f nv = (*itVertex)->data().getNormalArray();
        for (MFVec3f::const_iterator itNormal = nv.begin(); itNormal != nv.end(); ++itNormal)
          normalArray.push_back(*itNormal);
      }   
      
      // iterating the faces to create the normalIndexArray.
      MFInt32 normalIndexArray;
      
      for (MESH::Mesh::MFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
      {
        const MESH::Mesh::MEdge &me = (*itFace)->getEdges();
        for (MESH::Mesh::MEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
        {
          MESH::Mesh::Vertex *V = (*itEdge)->getFromVertex();
          
          V->data().getNormalMap();
          
          normalIndexArray.push_back(normalDecal[V->getIndex()] + V->data().getNormalMap()[*itFace]);
        }  
        normalIndexArray.push_back(-1);
      }
  
      I->setNormalIndex(normalIndexArray);
    }
    else
    {
      for (MESH::Mesh::MFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        normalArray.push_back((*itFace)->data().getNormal());
    }
    
    N->setVector(normalArray);
  }

  //////////////////////////////////////////////////////////////////////////////////// 
  // colors
  
  if (MM->data().hasColor())
  {
    if ((I->getColor() == 0) || (I->getColor()->getTypeName() == "ColorRGBA"))
    {
      // creating a new Node.
      ColorRGBA *C = new ColorRGBA();

      // adding parents.
      if (I->getColor() != 0)
      {
        MFAbstractNode parents = I->getColor()->getParents();
        for (MFAbstractNode::iterator it = parents.begin(); it != parents.end(); ++it)
          (*it)->addChild(C);
      }
      
      delete I->getColor();
    }
    
    MFColorRGBA colorArray;

    if (MM->data().getColorPerVertex())
    {
      // iterating the vertices to collect the informations relative to the colors
      vector<unsigned int> colorDecal;
      for (MESH::Mesh::MVertex::const_iterator itVertex = mv.begin(); itVertex != mv.end(); ++itVertex)
      {
        colorDecal.push_back(colorArray.size());
        
        MFColorRGBA nv = (*itVertex)->data().getColorArray();
        for (MFColorRGBA::const_iterator itColor = nv.begin(); itColor != nv.end(); ++itColor)
          colorArray.push_back(*itColor);
      }   
      
      // iterating the faces to create the colorIndexArray.
      MFInt32 colorIndexArray;
      
      for (MESH::Mesh::MFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
      {
        const MESH::Mesh::MEdge &me = (*itFace)->getEdges();
        for (MESH::Mesh::MEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
        {
          MESH::Mesh::Vertex *V = (*itEdge)->getFromVertex();
          
          V->data().getColorMap();
          
          colorIndexArray.push_back(colorDecal[V->getIndex()] + V->data().getColorMap()[*itFace]);
        }  
        colorIndexArray.push_back(-1);
      }
  
      I->setColorIndex(colorIndexArray);
    }
    else
    {
      for (MESH::Mesh::MFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
        colorArray.push_back((*itFace)->data().getColor());
    }
    
    static_cast<ColorRGBA *>(I->getColor())->setColor(colorArray);
  }
 
  //////////////////////////////////////////////////////////////////////////////////// 
  // texCoords
  
  if (MM->data().hasTexCoord())
  {
    TextureCoordinate *T = dynamic_cast<TextureCoordinate *>(I->getTexCoord());
    if (T == 0)
    {
      T = new TextureCoordinate();
      I->setTexCoord(T);
    }
    
    MFVec2f texCoordArray;

    // iterating the vertices to collect the informations relative to the normals
    vector<unsigned int> texCoordDecal;
    for (MESH::Mesh::MVertex::const_iterator itVertex = mv.begin(); itVertex != mv.end(); ++itVertex)
    {
      texCoordDecal.push_back(texCoordArray.size());
      
      MFVec2f nv = (*itVertex)->data().getTexCoordArray();
      for (MFVec2f::const_iterator itTexCoord = nv.begin(); itTexCoord != nv.end(); ++itTexCoord)
        texCoordArray.push_back(*itTexCoord);
    }   
    
    // iterating the faces to create the normalIndexArray.
    MFInt32 texCoordIndexArray;
    
    for (MESH::Mesh::MFace::const_iterator itFace = mf.begin(); itFace != mf.end(); ++itFace)  
    {
      const MESH::Mesh::MEdge &me = (*itFace)->getEdges();
      for (MESH::Mesh::MEdge::const_iterator itEdge = me.begin(); itEdge != me.end(); ++itEdge)
      {
        MESH::Mesh::Vertex *V = (*itEdge)->getFromVertex();
        
        V->data().getTexCoordMap();
        
        texCoordIndexArray.push_back(texCoordDecal[V->getIndex()] + V->data().getTexCoordMap()[*itFace]);
      }  
      texCoordIndexArray.push_back(-1);
    }
  
    I->setTexCoordIndex(texCoordIndexArray);
    T->setPoint(texCoordArray);
  }
}

}
}
